@tool
extends CompositorEffect
class_name PostProcessPixelOutline

var rd: RenderingDevice
var shader: RID
var pipeline: RID
var depth_sampler: RID
var normal_sampler: RID


func _init() -> void:
	effect_callback_type = EFFECT_CALLBACK_TYPE_PRE_TRANSPARENT
	rd = RenderingServer.get_rendering_device()
	RenderingServer.call_on_render_thread(_initialize_compute)


# System notifications, we want to react on the notification that
# alerts us we are about to be destroyed.
func _notification(what: int) -> void:
	if what == NOTIFICATION_PREDELETE:
		if shader.is_valid():
			# Freeing our shader will also free any dependents such as the pipeline!
			RenderingServer.free_rid(shader)


# Compile our shader at initialization.
func _initialize_compute() -> void:
	rd = RenderingServer.get_rendering_device()
	if not rd:
		return

	# Compile our shader.
	var shader_file                 = load("res://compositors/post_process_pixel_outline.glsl")
	var shader_spirv: RDShaderSPIRV = shader_file.get_spirv()

	# Create sampler state.
	var sampler_state: RDSamplerState = RDSamplerState.new()
	sampler_state.min_filter = RenderingDevice.SAMPLER_FILTER_NEAREST
	sampler_state.mag_filter = RenderingDevice.SAMPLER_FILTER_NEAREST
	depth_sampler = rd.sampler_create(sampler_state)
	normal_sampler = rd.sampler_create(sampler_state)

	shader = rd.shader_create_from_spirv(shader_spirv)
	if shader.is_valid():
		pipeline = rd.compute_pipeline_create(shader)


# Called by the rendering thread every frame.
func _render_callback(p_effect_callback_type: EffectCallbackType, p_render_data: RenderData) -> void:
	if rd and p_effect_callback_type == EFFECT_CALLBACK_TYPE_PRE_TRANSPARENT and pipeline.is_valid():
		# Get our render scene buffers object, this gives us access to our render buffers.
		# Note that implementation differs per renderer hence the need for the cast.
		var render_scene_buffers: RenderSceneBuffersRD = p_render_data.get_render_scene_buffers()
		if render_scene_buffers:
			# Get our render size, this is the 3D render resolution!
			var size: Vector2i = render_scene_buffers.get_internal_size()
			if size.x == 0 and size.y == 0:
				return

			# We can use a compute shader here.
			@warning_ignore("integer_division")
			var x_groups := (size.x - 1) / 8 + 1
			@warning_ignore("integer_division")
			var y_groups := (size.y - 1) / 8 + 1
			var z_groups := 1

			var render_size = render_scene_buffers.get_internal_size()

			# Create push constant.
			# Must be aligned to 16 bytes and be in the same order as defined in the shader.
			var push_constant := PackedFloat32Array([
			size.x,
			size.y,
			render_size.x,
			render_size.y,
			])

			# Loop through views just in case we're doing stereo rendering. No extra cost if this is mono.
			var view_count: int = render_scene_buffers.get_view_count()
			for view in view_count:
				var input_image: RID = render_scene_buffers.get_color_layer(view)
				var input_uniform    = RDUniform.new()
				input_uniform.uniform_type = RenderingDevice.UNIFORM_TYPE_IMAGE
				input_uniform.binding = 0
				input_uniform.add_id(input_image)

				var depth_image: RID = render_scene_buffers.get_depth_layer(view)
				var depth_uniform    = RDUniform.new()
				depth_uniform.uniform_type = RenderingDevice.UNIFORM_TYPE_SAMPLER_WITH_TEXTURE
				depth_uniform.binding = 0
				depth_uniform.add_id(depth_sampler)
				depth_uniform.add_id(depth_image)

				var normal_image: RID = render_scene_buffers.get_texture("forward_clustered", "normal_roughness")
				var normal_uniform    = RDUniform.new()
				normal_uniform.uniform_type = RenderingDevice.UNIFORM_TYPE_SAMPLER_WITH_TEXTURE
				normal_uniform.binding = 0
				normal_uniform.add_id(normal_sampler)
				normal_uniform.add_id(normal_image)

				var color_uniform_set  = UniformSetCacheRD.get_cache(shader, 0, [input_uniform])
				var depth_uniform_set  = UniformSetCacheRD.get_cache(shader, 1, [depth_uniform])
				var normal_uniform_set = UniformSetCacheRD.get_cache(shader, 2, [normal_uniform])
			
				var compute_list := rd.compute_list_begin()
				rd.compute_list_bind_compute_pipeline(compute_list, pipeline)
				rd.compute_list_bind_uniform_set(compute_list, color_uniform_set, 0)
				rd.compute_list_bind_uniform_set(compute_list, depth_uniform_set, 1)
				rd.compute_list_bind_uniform_set(compute_list, normal_uniform_set, 2)
				rd.compute_list_set_push_constant(compute_list, push_constant.to_byte_array(), push_constant.size() * 4)
				rd.compute_list_dispatch(compute_list, x_groups, y_groups, z_groups)
				rd.compute_list_end()